from .. import DATA_TYPES, ET, BaseCodec, CodecList


class FieldCodec(BaseCodec):
    """The base class for a Field.
    
    Attributes:
        data_type (str): The data type from a supported list.
        name (str): The unique Field name.
        description (str): Optional description.
        optional (bool): Optional indication the field is optional.

    """
    def __init__(self,
                 name: str,
                 data_type: str,
                 description: str = None,
                 optional: bool = False) -> None:
        """Instantiates the base field.
        
        Args:
            name: The field name must be unique within a Message.
            data_type: The data type represented within the field.
            description: (Optional) Description/purpose of the field.
            optional: (Optional) Indicates if the field is mandatory.
            
        """
        super().__init__(name, description)
        if data_type not in DATA_TYPES:
            raise ValueError(f'Invalid data type {data_type}')
        self._data_type = data_type
        self._optional = optional
    
    @property
    def data_type(self) -> str:
        return self._data_type

    @property
    def optional(self) -> bool:
        return self._optional
    
    @optional.setter
    def optional(self, value: bool):
        if not value or not isinstance(value, bool):
            value = False
        self._optional = value

    @property
    def bits(self) -> int:
        """Must be subclassed."""
        raise NotImplementedError('Subclass must define bits')

    def __repr__(self) -> str:
        rep = {}
        for name in dir(self):
            if name.startswith(('__', '_')):
                continue
            attr = getattr(self, name)
            if not callable(attr):
                rep[name] = attr
        return repr(rep)
    
    def _base_xml(self) -> ET.Element:
        """The default XML template for a Field."""
        xsi_type = DATA_TYPES[self.data_type]
        xmlfield = ET.Element('Field', attrib={
            '{http://www.w3.org/2001/XMLSchema-instance}type': xsi_type
        })
        name = ET.SubElement(xmlfield, 'Name')
        name.text = self.name
        if self.description:
            description = ET.SubElement(xmlfield, 'Description')
            description.text = str(self.description)
        if self.optional:
            optional = ET.SubElement(xmlfield, 'Optional')
            optional.text = 'true'
        return xmlfield
    
    def decode(self, *args, **kwargs):
        """Must be subclassed."""
        raise NotImplementedError('Subclass must define decode')
    
    def encode(self, *args, **kwargs):
        """Must be subclassed."""
        raise NotImplementedError('Subclass must define encode')
    
    def xml(self, *args, **kwargs):
        """Must be subclassed."""
        raise NotImplementedError('Subclass must define xml structure')


class Fields(CodecList):
    """The list of Fields defining a Message or ArrayElement."""
    def __init__(self, fields: 'list[FieldCodec]' = None):
        super().__init__(codec_cls=FieldCodec)
        if fields is not None:
            for field in fields:
                self.add(field)
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Fields):
            return NotImplemented
        if len(self) != len(other):
            return False
        for i, field in enumerate(self):
            if field != other[i]:
                return False
        return True



from .. import ET
from .base_field import FieldCodec
from .helpers import optimal_bits


class EnumField(FieldCodec):
    """An enumerated field sends an index over-the-air representing a string."""
    def __init__(self,
                 name: str,
                 items: 'list[str]',
                 size: int,
                 description: str = None,
                 optional: bool = False,
                 default: int = None,
                 value: int = None) -> None:
        """Instantiates a EnumField.
        
        Args:
            name: The field name must be unique within a Message.
            items: A list of strings (indexed from 0).
            size: The number of *bits* used to encode the index over-the-air.
            description: An optional description/purpose for the field.
            optional: Indicates if the field is optional in the Message.
            default: A default value for the enum.
            value: Optional value to set during initialization.

        """
        super().__init__(name=name,
                         data_type='enum',
                         description=description,
                         optional=optional)
        if (not isinstance(items, list) or
            not all(isinstance(item, str) for item in items)):
            raise ValueError('Items must a list of strings')
        self._items = items
        min_size = optimal_bits((0, len(items) - 1))
        if not isinstance(size, int) or size < min_size:
            raise ValueError(f'Size must be integer greater than {min_size}')
        self._size = size
        if default is not None:
            if isinstance(default, str):
                if default not in items:
                    raise ValueError(f'{default} not found in items')
                self._default = items.index(default)
            elif isinstance(default, int):
                if default not in range(0, len(items)):
                    raise ValueError('Invalid default not in range of items')
                self._default = default
        else:
            self._default = None
        if value is not None:
            if value not in items:
                raise ValueError(f'{value} not in items')
            self._value = value
        else:
            self._value = None
    
    def _validate_enum(self, v: 'int|str') -> 'int|None':
        if v is not None:
            if isinstance(v, str):
                if v not in self.items:
                    raise ValueError(f'Invalid value {v}')
                for index, item in enumerate(self.items):
                    if item == v:
                        return index
            elif isinstance(v, int):
                if v < 0 or v >= len(self.items):
                    raise ValueError(f'Invalid enum index {v}')
            else:
                raise ValueError(f'Invalid value {v}')
        return v

    @property
    def items(self):
        return self._items
    
    @items.setter
    def items(self, l: list):
        if not isinstance(l, list) or not all(isinstance(x, str) for x in l):
            raise ValueError('Items must be a list of strings')
        self._items = l

    @property
    def default(self) -> str:
        if self._default is None:
            return None
        return self.items[self._default]
    
    @default.setter
    def default(self, v: 'int|str'):
        self._default = self._validate_enum(v)

    @property
    def value(self) -> str:
        if self._value is None:
            if self.default is not None:
                return self.default
            return None
        return self.items[self._value]
    
    @value.setter
    def value(self, v: 'int|str'):
        self._value = self._validate_enum(v)

    @property
    def size(self) -> int:
        """The size of the field in bits."""
        return self._size
    
    @size.setter
    def size(self, v: int):
        if not isinstance(v, int) or v < 1:
            raise ValueError('Size must be integer greater than zero')
        minimum_bits = optimal_bits((0, len(self.items)))
        if v < minimum_bits:
            raise ValueError(f'Size must be at least {minimum_bits}'
                             ' to support item count')
        self._size = v

    @property
    def bits(self) -> int:
        """The size of the field in bits."""
        bits = self.size if self._value is not None else 0
        return bits + (1 if self.optional else 0)
    
    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None:
            raise ValueError(f'No value configured in EnumField {self.name}')
        _format = f'0{self.size}b'
        binstr = format(self.items.index(self.value), _format)
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        self.value = int(binary_str[:self.size], 2)
        return self.size

    def xml(self) -> ET.Element:
        """Returns the Enum XML definition for a Message Definition File."""
        # Size must come after Items for Inmarsat V1 parser
        xmlfield = self._base_xml()
        items = ET.SubElement(xmlfield, 'Items')
        for string in self.items:
            item = ET.SubElement(items, 'string')
            item.text = str(string)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        return xmlfield



from copy import deepcopy

from .. import ET
from .base_field import FieldCodec, Fields
from .helpers import decode_field_length, encode_field_length


class ArrayField(FieldCodec):
    """An Array Field provides a list where each element is a set of Fields.
    
    Attributes:
        name (str): The name of the field instance.
        size (int): The maximum number of elements allowed.
        fields (Fields): A list of Field types comprising each ArrayElement
        description (str): An optional description of the array/use.
        optional (bool): Indicates if the array is optional in the Message
        fixed (bool): Indicates if the array is always the fixed `size`
        elements (list): The enumerated list of ArrayElements

    """
    def __init__(self,
                 name: str,
                 size: int,
                 fields: Fields,
                 description: str = None,
                 optional: bool = False,
                 fixed: bool = False,
                 elements: 'list[Fields]' = []) -> None:
        """Initializes an ArrayField instance.
        
        Args:
            name: The unique field name within the Message.
            size: The maximum number of elements allowed.
            fields: The list of Field types comprising each element.
            description: An optional description/purpose of the array.
            optional: Indicates if the array is optional in the Message.
            fixed: Indicates if the array is always the fixed `size`.
            elements: Option to populate elements of Fields during instantiation.

        """
        super().__init__(name=name,
                         data_type='array',
                         description=description,
                         optional=optional)
        self._size = size
        self._fixed = fixed
        self._fields = fields
        self._elements = elements or []
    
    @property
    def size(self) -> int:
        """The maximum number of array elements."""
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0')
        self._size = value
    
    @property
    def fixed(self) -> bool:
        """Indicates if the array is a fixed size (padded with defaults)."""
        return self._fixed

    @property
    def fields(self) -> Fields:
        """The set of `FieldCodec`s that make up each array element."""
        return self._fields

    @fields.setter
    def fields(self, fields: Fields):
        if not isinstance(fields, Fields):
            raise ValueError('Invalid Fields definition for ArrayField')
        self._fields = fields

    @property
    def elements(self) -> 'list[Fields]':
        """The list of elements (field sets) in the array."""
        return self._elements
    
    @elements.setter
    def elements(self, elements: 'list[Fields]'):
        if (not isinstance(elements, list) or 
            not all(isinstance(item, Fields) for item in elements)):
            raise ValueError('Elements must be a list of grouped Fields')
        for fields in elements:
            # assert isinstance(fields, Fields)
            for index, field in enumerate(fields):
                assert isinstance(field, FieldCodec)
                if (field.name != self.fields[index].name):
                    raise ValueError(f'fields[{index}].name'
                                     f' expected {self.fields[index].name}'
                                     f' got {field.name}')
                if (field.data_type != self.fields[index].data_type):
                    raise ValueError(f'fields[{index}].data_type'
                                     f' expected {self.fields[index].data_type}'
                                     f' got {field.data_type}')
                #TODO: validate non-optional fields have value/elements
                if (not field.optional and
                    not isinstance(field, ArrayField) and
                    field.value is None):
                    raise ValueError(f'fields[{index}].value missing')
                try:
                    self._elements[index] = fields
                except IndexError:
                    self._elements.append(fields)

    @property
    def bits(self) -> int:
        """The size of the array in bits."""
        bits = 0
        for field in self.fields:
            assert isinstance(field, FieldCodec)
            bits += field.bits
        return bits + (1 if self.optional else 0)
    
    def _valid_element(self, element: Fields) -> bool:
        for i, field in enumerate(self.fields):
            assert isinstance(field, FieldCodec)
            e_field = element[i]
            assert isinstance(e_field, FieldCodec)
            if e_field.name != field.name:
                raise ValueError(f'element field name {e_field.name}'
                                 f' does not match {field.name}')
            if e_field.data_type != field.data_type:
                raise ValueError(f'element field data_type {e_field.data_type}'
                                 f' does not match {field.data_type}')
            if e_field.optional != field.optional:
                raise ValueError(f'element optional {e_field.optional}'
                                 f' does not match {field.optional}')
            if (hasattr(field, 'fixed') and
                hasattr(e_field, 'fixed') and
                e_field.fixed != field.fixed):
                raise ValueError(f'element fixed {e_field.fixed}'
                                 f' does not match {field.fixed}')
            if (hasattr(field, 'size') and
                hasattr(e_field, 'size') and
                e_field.size != field.size):
                raise ValueError(f'element size {e_field.size}'
                                 f' does not match {field.size}')
        return True

    def append(self, element: Fields):
        """Adds the array element to the list of elements."""
        if not isinstance(element, Fields):
            raise ValueError('Invalid element definition must be Fields')
        if not self._valid_element(element):
            raise ValueError('Invalid element definition'
                             f' - requires {self.fields}')
        for i, field in enumerate(element):
            assert isinstance(field, FieldCodec)
            if (hasattr(field, 'description') and
                field.description != self.fields[i].description):
                element[i].description = self.fields[i].description
            if hasattr(field, 'value') and field.value is None:
                element[i].value = self.fields[i].default
        self._elements.append(element)

    def new_element(self) -> Fields:
        """Returns an empty element at the end of the elements list."""
        new_index = len(self._elements)
        new_fields = deepcopy(self.fields)
        self.append(Fields(new_fields))
        return self.elements[new_index]

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if len(self.elements) == 0:
            raise ValueError('No elements to encode')
        binstr = ''
        for element in self.elements:
            for field in element:
                binstr += field.encode()
        if not self.fixed:
            binstr = encode_field_length(len(self.elements)) + binstr
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        if self.fixed:
            length = self.size
            bit_index = 0
        else:
            (length, bit_index) = decode_field_length(binary_str)
        for index in range(0, length):
            fields = Fields(self.fields)
            for field in fields:
                if field.optional:
                    if binary_str[bit_index] == '0':
                        bit_index += 1
                        continue
                    bit_index += 1
                bit_index += field.decode(binary_str[bit_index:])
            try:
                self._elements[index] = fields
            except IndexError:
                self._elements.append(fields)
        return bit_index

    def xml(self) -> ET.Element:
        """Returns the Array XML definition for a Message Definition File."""
        # Size must come after Fields for Inmarsat V1 parser
        xmlfield = self._base_xml()
        if self.fixed:
            default = ET.SubElement(xmlfield, 'Fixed')
            default.text = 'true'
        fields = ET.SubElement(xmlfield, 'Fields')
        for field in self.fields:
            assert isinstance(field, FieldCodec)
            fields.append(field.xml())
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        return xmlfield



XML_NAMESPACE = {
    'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
    'xsd': 'http://www.w3.org/2001/XMLSchema'
}

DATA_TYPES = {
    'bool': 'BooleanField',
    'int_8': 'SignedIntField',
    'uint_8': 'UnsignedIntField',
    'int_16': 'SignedIntField',
    'uint_16': 'UnsignedIntField',
    'int_32': 'SignedIntField',
    'uint_32': 'UnsignedIntField',
    'int_64': 'SignedIntField',
    'uint_64': 'UnsignedIntField',
    'float': 'DataField',
    'double': 'DataField',
    'string': 'StringField',
    'data': 'DataField',
    'enum': 'EnumField',
    'array': 'ArrayField',
}

SIN_RANGE = (16, 255)



"""Codec functions for IDP Common Message Format supported by Inmarsat MGS.

Also supported on ORBCOMM IGWS.

*** common_mdf DEPRECATED - use common_message_format instead ***

"""
from .common_message_format import *


from .array_field import ArrayField
from .boolean_field import BooleanField
from .data_field import DataField
from .enum_field import EnumField
from .integer_field import SignedIntField, UnsignedIntField
from .string_field import StringField



class BaseCodec:
    def __init__(self, name: str, description: str = None) -> None:
        if not name or name.strip() == '':
            raise ValueError('Invalid name must be non-empty')
        self._name = name
        self._description = description
    
    @property
    def name(self) -> str:
        return self._name
    
    @property
    def description(self) -> str:
        return self._description
    
    @description.setter
    def description(self, value: str):
        if value is not None and not isinstance(value, str) and not value:
            raise ValueError('Description must be non-empty string or None')
        self._description = value

    def _attribute_equivalence(self,
                               other: object,
                               exclude: "list[str]" = None) -> bool:
        """Indicates attribute equivalence, with optional exceptions."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        for attr, val in self.__dict__.items():
            if exclude is not None and attr in exclude:
                continue
            if not hasattr(other, attr) or val != other.__dict__[attr]:
                return False
        return True
    
    def __eq__(self, other: object) -> bool:
        return self._attribute_equivalence(other)


class CodecList(list):
    """Base class for a specific object type list.
    
    Used for Fields, Messages, Services.

    Attributes:
        codec_cls: The object type the list is comprised of.

    """
    def __init__(self, codec_cls: BaseCodec):
        super().__init__()
        self.list_type = codec_cls

    def add(self, codec: BaseCodec) -> bool:
        """Add an object to the end of the list.

        Args:
            obj (object): A valid object according to the list_type
        
        Raises:
            ValueError if there is a duplicate or invalid name,
                invalid value_range or unsupported data_type
        """
        if not isinstance(codec, self.list_type):
            raise ValueError(f'Invalid {self.list_type} definition')
        for o in self:
            if o.name == codec.name:
                raise ValueError(f'Duplicate {self.list_type}'
                                 f' name {codec.name} found')
        self.append(codec)
        return True

    def __getitem__(self, n: 'str|int') -> BaseCodec:
        """Retrieves an object by name or index.
        
        Args:
            n: The object name or list index
        
        Returns:
            object

        """
        if isinstance(n, str):
            for o in self:
                if o.name == n:
                    return o
            raise ValueError(f'{self.list_type} name {n} not found')
        return super().__getitem__(n)

    def __setitem__(self, n: 'str|int', value):
        if isinstance(n, str):
            for o in self:
                if o.name == n:
                    o.value = value
                    break
        else:
            super().__setitem__(n, value)

    def delete(self, name: str) -> bool:
        """Delete an object from the list by name.
        
        Args:
            name: The name of the object.

        Returns:
            boolean: success
        """
        for o in self:
            if o.name == name:
                self.remove(o)
                return True
        return False
